﻿using PalmSens;
using PalmSens.Analysis;
using PalmSens.Comm;
using PalmSens.Core.Simplified.Data;
using PalmSens.Core.Simplified.WPF;
using PalmSens.Devices;
using PalmSens.Fitting;
using PalmSens.Fitting.Models;
using PalmSens.Fitting.Models.Circuits;
using PalmSens.Plottables;
using PalmSens.Techniques;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

namespace PSSDKMultiEISFitExample
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        public MainWindow()
        {
            InitializeComponent();

            InitPlot(); //Resets and initiates the plot control
            LoadData(); //Loads data from example .pssession

            _fitProgressBars = new ProgressBar[] { progressBar0, progressBar1, progressBar2 };
            _iterationLabels = new Label[] { lblIterations0, lblIterations1, lblIterations2 };
        }

        /// <summary>
        /// The fit progress bars
        /// </summary>
        private ProgressBar[] _fitProgressBars;

        /// <summary>
        /// The fit iteration counter labels
        /// </summary>
        private Label[] _iterationLabels;

        /// <summary>
        /// The active SimpleMeasurement
        /// </summary>
        private SimpleMeasurement[] _simpleMeasurements = null;

        /// <summary>
        /// The fit result SimpleCurve for each fitted circuit model
        /// </summary>
        private Dictionary<CircuitModel, SimpleCurve[]> _fitResultCurves = new Dictionary<CircuitModel, SimpleCurve[]>();

        /// <summary>
        /// Dictionary with circuit model class defining the model that will be used for the equivalent circuit fit in the corresponding fit algorithm class
        /// </summary>
        private Dictionary<FitAlgorithm, CircuitModel> _circuitModels = new Dictionary<FitAlgorithm, CircuitModel>();

        /// <summary>
        /// The fitted measurents and their equivalent circuit models defined in the circuit model class
        /// </summary>
        private Dictionary<CircuitModel, SimpleMeasurement> _fittedMeasurements = new Dictionary<CircuitModel, SimpleMeasurement>();

        /// <summary>
        /// The fit algorithms and their corresponding progress classes
        /// </summary>
        private Dictionary<FitProgress, FitAlgorithm> _fitAlgorithms = new Dictionary<FitProgress, FitAlgorithm>();

        /// <summary>
        /// Index defining which ui elements to update
        /// </summary>
        private Dictionary<FitProgress, int> _progressUI = new Dictionary<FitProgress, int>();

        /// <summary>
        /// Initializes the plot control.
        /// </summary>
        private void InitPlot()
        {
            plot.ClearAll(); //Clear all curves and data from plot
            //Set the Axis labels
            plot.XAxisLabel = "Log Frequency (Hz)";
            plot.XAxisType = SDKPlot.AxisType.Logarithmic;
            plot.YAxisLabel = "Log Z (Ω)";
            plot.YAxisType = SDKPlot.AxisType.Logarithmic;
            plot.YAxisSecondaryLabel = "-Phase (°)";
            plot.YAxisSecondaryType = SDKPlot.AxisType.Linear;
            plot.AddData("", new double[0], new double[0]); //Add a empty data array to draw an empty plot
        }

        /// <summary>
        /// Loads the data from a pssession file.
        /// </summary>
        private void LoadData()
        {
            //Model settings for EIS measurements in file
            string[] cdcs = new string[] { "R(RC)(RC)", "R(RC)", "R(RQ)" };
            double[][] initialParams = new double[][] {
                new double[] { 10, 1.2e4, 1e-7, 2.5e4, 5e-7 },
                new double[] { 2000, 9000, 1e-7 },
                new double[] { 3, 40, 1e-4, 0.7 }
            };

            _simpleMeasurements = SimpleLoadSaveFunctions.LoadMeasurements("MultipleEISExamples.pssession").ToArray();

            //Add bode curves to plot
            for (int i = 0; i < 3; i++)
            {
                plot.AddSimpleCurve(_simpleMeasurements[i].NewSimpleCurve(PalmSens.Data.DataArrayType.Frequency, PalmSens.Data.DataArrayType.Z, "Z Abs")[0]);
                plot.AddSimpleCurve(_simpleMeasurements[i].NewSimpleCurve(PalmSens.Data.DataArrayType.Frequency, PalmSens.Data.DataArrayType.Phase, "-Phase")[0], true);

                //Build the model to fit on the data
                CircuitModel model = BuildEquivalentCircuitModel(_simpleMeasurements[i].Measurement.EISdata[0], cdcs[i], initialParams[i]);
                _fittedMeasurements.Add(model, _simpleMeasurements[i]);
            }

            //Fix the first parameter of the second model
            CircuitModel model2 = _fittedMeasurements.Keys.ToArray()[1];
            Parameters modelParameters = model2.InitialParameters;
            modelParameters[0].Fixed = true;

            //Set parameter bounds for third model
            CircuitModel model3 = _fittedMeasurements.Keys.ToArray()[2];
            modelParameters = model3.InitialParameters;
            modelParameters[0].MinValue = 1; modelParameters[0].MaxValue = 10;
            modelParameters[1].MinValue = 20; modelParameters[1].MaxValue = 100;
            modelParameters[2].MinValue = 1e-5; modelParameters[2].MaxValue = 1e-3;
            modelParameters[3].MinValue = .6; modelParameters[3].MaxValue = 1;

            btnFit.IsEnabled = true;
        }

        /// <summary>
        /// Handles the Click event of the BtnFit control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void BtnFit_Click(object sender, RoutedEventArgs e)
        {
            btnFit.IsEnabled = false;
            await FitEquivalentCircuits(); //Fits a specified equivalent circuit to the measured data
            btnFit.IsEnabled = true;
        }

        /// <summary>
        /// Handles the Click event of the BtnCancel control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private void BtnCancel_Click(object sender, RoutedEventArgs e)
        {
            foreach(FitProgress progress in _fitAlgorithms.Keys)
                progress.Cancel(); //Cancels the active fit
        }

        /// <summary>
        /// Fits the equivalent circuit.
        /// </summary>
        /// <exception cref="System.NotImplementedException"></exception>
        private async Task FitEquivalentCircuits()
        {
            float[][] ranges = new float[][] { new float[2] { 0.1f, 1e6f }, new float[2] { 0.01f, 1e5f }, new float[2] { 2, 100 } }; //fit range in Hz

            //Clear previous fit info
            _circuitModels.Clear();
            _progressUI.Clear();
            _fitAlgorithms.Clear();

            for (int i = 0; i < 3; i++)
            {
                EISData data = _simpleMeasurements[i].Measurement.EISdata[0];
                CircuitModel model = _fittedMeasurements.Keys.ToArray()[i];
                FitOptionsCircuit fitOptions = InitFitOptions(model, data); //Define the options for the fit
                fitOptions.SelectedDataPoints = GetFitRange(_simpleMeasurements[i], ranges[i]);
                FitAlgorithm fitAlgorithm = FitAlgorithm.FromAlgorithm(fitOptions); //Set up the instance of the fit class
                FitProgress fitProgress = new FitProgress(); //Create an instance of the peak detection progress reporter (this can also be used to cancel the peak detection)
                fitProgress.ProgressChanged += FitProgress_ProgressChanged;

                _circuitModels.Add(fitAlgorithm, model);
                _fitAlgorithms.Add(fitProgress, fitAlgorithm);
                _progressUI.Add(fitProgress, i);
            }

            List<Task> fitTasks = new List<Task>();
            foreach (KeyValuePair<FitProgress, FitAlgorithm> keyValuePair in _fitAlgorithms)
                fitTasks.Add(keyValuePair.Value.ApplyFitCircuitAsync(keyValuePair.Key));

            await Task.WhenAll(fitTasks);
        }

        /// <summary>
        /// Builds the equivalent circuit model.
        /// </summary>
        /// <returns></returns>
        private CircuitModel BuildEquivalentCircuitModel(EISData data, string cdc, double[] initialParams)
        {
            CircuitModel circuitModel = new CircuitModel();
            circuitModel.SetEISdata(data); //Sets reference to measured data
            circuitModel.SetCircuit(cdc); //Sets the circuit defined in the CDC code string, in this case a Randles circuit
            
            //Setting the initial parameters is recommended to ensure a good quality fit and reduce to the time it takes to find the fit
            circuitModel.SetInitialParameters(initialParams);

            return circuitModel;
        }

        /// <summary>
        /// Initializes the fit options.
        /// </summary>
        /// <param name="circuitModel">The circuit model.</param>
        /// <returns></returns>
        private FitOptionsCircuit InitFitOptions(CircuitModel circuitModel, EISData data)
        {
            FitOptionsCircuit fitOptions = new FitOptionsCircuit();
            fitOptions.Model = circuitModel;
            fitOptions.RawData = data; //The measured data to fit onto
            fitOptions.SelectedAlgorithm = Algorithm.LevenbergMarquardt;
            fitOptions.MaxIterations = 100; //The maximum number of iterations, 500 by default
            fitOptions.MinimumDeltaErrorTerm = 1e-9; //The minimum delta in the error term (sum of squares difference between model and data), default is 1e-9      
            fitOptions.EnableMinimunDeltaErrorTerm = true; //Enable the minimum delta error as a stop condition within the fit algorithm, default is true
            fitOptions.MinimumDeltaParameters = 1e-12; //The minimum delta parameter step size, default is 1e-12
            fitOptions.EnableMinimunDeltaParameters = true; //Enable the minimum delta parameter step size as a stop condition within the fit algorithm, default is true
            fitOptions.Lambda = 1e-2; //The starting value for the Levenberg Marquardt scaling factor, default is 1e-2
            fitOptions.LambdaFactor = 10; //The scaling value for the Levenberg Marquardt scaling factor, default is 10

            return fitOptions;
        }

        /// <summary>
        /// Sets the fit range.
        /// </summary>
        /// <param name="simpleMeasurement">The simple measurement.</param>
        /// <param name="range">The fit range in Hz.</param>
        /// <returns>A bool indicating the selected range of the fit</returns>
        /// <exception cref="NotImplementedException"></exception>
        private bool[] GetFitRange(SimpleMeasurement simpleMeasurement, float[] range)
        {
            double[] frequencies = simpleMeasurement.GetRawData(PalmSens.Data.DataArrayType.Frequency);
            bool[] fitRange = new bool[frequencies.Length];
            for (int i = 0; i < frequencies.Length; i++)
                fitRange[i] = frequencies[i] > range[0] && frequencies[i] < range[1];
            return fitRange;
        }

        /// <summary>
        /// Raised when fit progress changed.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The e.</param>
        /// <exception cref="NotImplementedException"></exception>
        private void FitProgress_ProgressChanged(object sender, FitProgressUpdate e)
        {
            FitProgress fitProgress = sender as FitProgress;
            int index = _progressUI[fitProgress];
            switch (e.Progress)
            {
                case EnumFitProgress.Started:
                    lbResults.Items.Add("Started circuit fit.");
                    _iterationLabels[index].Content = "";
                    _fitProgressBars[index].IsIndeterminate = btnCancel.IsEnabled = true;
                    break;
                case EnumFitProgress.FitIterated:
                    _iterationLabels[index].Content = e.NIterations.ToString();
                    break;
                case EnumFitProgress.Cancelled:
                case EnumFitProgress.Finished:
                    _fitProgressBars[index].IsIndeterminate = btnCancel.IsEnabled = false;
                    _iterationLabels[index].Content = (e.Result.NIterations - 1).ToString(); //Excludes initial guess as iteration
                    SetFitResults(sender as FitProgress, e.Result);
                    lbResults.Items.Add("Circuit fit completed.");
                    fitProgress.ProgressChanged -= FitProgress_ProgressChanged;
                    _fitAlgorithms.Remove(fitProgress);
                    break;
            }
        }

        /// <summary>
        /// Sets the fit results.
        /// </summary>
        /// <param name="result">The result.</param>
        private void SetFitResults(FitProgress progress, FitResult result)
        {
            CircuitModel model = _circuitModels[_fitAlgorithms[progress]];
            SimpleMeasurement measurement = _fittedMeasurements[model];

            //Add results to console
            lbResults.Items.Add($"Fit ({measurement.Title})");
            for (int i = 0; i < model.InitialParameters.Count; i++)
            {
                Parameter p = model.InitialParameters[i];
                lbResults.Items.Add($"{p.Name} = {result.FinalParameters[i].ToString("0.00E+000")}");
                lbResults.Items.Add($"%error = {result.ParameterSDs[i].ToString("0.00E+000")} %");
            }
            switch (result.ExitCode)
            {
                case ExitCodes.MinimumDeltaErrorTerm:
                    lbResults.Items.Add("Reached min delta error sum squares");
                    break;
                case ExitCodes.MinimumDeltaParameters:
                    lbResults.Items.Add("Reached min delta parameter step size");
                    break;
                case ExitCodes.MaxIterations:
                    lbResults.Items.Add("Reached max iterations");
                    break;
                case ExitCodes.HessianNonPositive:
                    lbResults.Items.Add("Error determing step for next iteration");
                    break;
            }

            //Remove previous fit results from plot
            if (_fitResultCurves.ContainsKey(model))
                foreach (SimpleCurve sc in _fitResultCurves[model])
                    if (plot.ContainsSimpleCurve(sc))
                        plot.RemoveSimpleCurve(sc);

            //Generate new simple curves based on fit results
            model.SetInitialParameters(result.FinalParameters); //Update the initial parameters with fit results
            SimpleCurve[] fitResultCurves = new SimpleCurve[2];
            fitResultCurves[0] = new SimpleCurve(model.GetCurveZabsOverFrequency(false)[0], measurement);
            fitResultCurves[0].Title = "Z Abs Fit";
            fitResultCurves[1] = new SimpleCurve(model.GetCurvePhaseOverFrequency(false)[0], measurement);
            fitResultCurves[1].Title = "-Phase Fit";

            if (_fitResultCurves.ContainsKey(model))
                _fitResultCurves[model] = fitResultCurves;
            else
                _fitResultCurves.Add(model, fitResultCurves);

            //Add results to plot   
            plot.AddSimpleCurve(fitResultCurves[0], false, false);
            plot.AddSimpleCurve(fitResultCurves[1], true, true);
        }
    }
}
