﻿using PalmSens;
using PalmSens.Comm;
using PalmSens.Core.Simplified.Data;
using PalmSens.Devices;
using PalmSens.Techniques;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

namespace PSSDKMultiplexerExample
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        public MainWindow()
        {
            InitializeComponent();

            DataContext = this; //Set the DataContext, needed to bind the measured data to the DataGrid control
            InitCAMethod(); //Create the chronoamperometry method that defines the measurement parameters
            DiscoverConnectedDevices(); //Populate the connected device combobox control

            //Init channel list for the selected channel combobox control
            Channels = new ObservableCollection<int>();
            for (int i = 1; i <= 16; i++)
                Channels.Add(i);
        }

        /// <summary>
        /// The instance of method class containing the Chronoamperometry parameters
        /// </summary>
        private AmperometricDetection _methodCA;

        /// <summary>
        /// The instance of method class containing the Chronoamperometry parameters 
        /// and mulitplexer settings for sequential measurements on different channels
        /// </summary>
        private AmperometricDetection _methodCASequential;

        /// <summary>
        /// The instance of method class containing the Chronoamperometry parameters 
        /// and mulitplexer settings for alternating measurements on different channels
        /// </summary>
        private AmperometricDetection _methodCAAlternating;

        /// <summary>
        /// The connected PalmSens & EmStat devices
        /// </summary>
        private Device[] _connectedDevices = new Device[0];

        /// <summary>
        /// The active SimpleMeasurement
        /// </summary>
        private SimpleMeasurement _activeMeasurement = null;

        /// <summary>
        /// Collection of multiplexer channels binded to the channel combobox control
        /// </summary>
        public ObservableCollection<int> Channels { get; private set; }

        /// <summary>
        /// The collection of measured data points that is binded to the datagrid.
        /// </summary>
        public ObservableCollection<DataPoint> DataPoints
        {
            get { return _dataPoints; }
            set { _dataPoints = value; }
        }

        /// <summary>
        /// The measured data points
        /// </summary>
        private ObservableCollection<DataPoint> _dataPoints = new ObservableCollection<DataPoint>();

        /// <summary>
        /// Initializes the CA method.
        /// </summary>
        private void InitCAMethod()
        {
            _methodCA = new AmperometricDetection(); //Create a new chronoamperometry method with the default settings
            _methodCA.Potential = 1f; //Sets the potential at 1V
            _methodCA.IntervalTime = 0.1f; //Sets the sampling interval time at 0.1s
            _methodCA.RunTime = 3f; //Sets the measurement runtime at 3s

            _methodCA.EquilibrationTime = 1f; //Equilabrates the cell at the defined potential for 1 second before starting the measurement
            _methodCA.Ranging.StartCurrentRange = new CurrentRange(CurrentRanges.cr1uA); //Starts equilabration in the 1µA current range
            _methodCA.Ranging.MinimumCurrentRange = new CurrentRange(CurrentRanges.cr10nA); //Min current range 10nA
            _methodCA.Ranging.MaximumCurrentRange = new CurrentRange(CurrentRanges.cr1mA); //Max current range 1mA

            //Sequential multiplexer measurement settings
            _methodCASequential = new AmperometricDetection(); //Create a new chronoamperometry method with the default settings
            _methodCASequential.Potential = 1f; //Sets the potential at 1V
            _methodCASequential.IntervalTime = 0.1f; //Sets the sampling interval time at 0.1s
            _methodCASequential.RunTime = 1f; //Sets the measurement runtime at 1s
            _methodCASequential.MuxMethod = MuxMethod.Sequentially; //Set the type of multiplexer measurement
            //Create an array of booleans indicating on which channels you would like to measure
            //For a 16 channel mux you can create a bool array for 16 channels
            bool[] measureOnChannels1 = new bool[8];  //Define bool array for 8 channel multiplexer
            //Specify on which channels to measure            
            measureOnChannels1[0] = false;
            measureOnChannels1[1] = true;
            measureOnChannels1[2] = true;
            measureOnChannels1[3] = false;
            measureOnChannels1[4] = true;
            measureOnChannels1[5] = false;
            measureOnChannels1[6] = false;
            measureOnChannels1[7] = false;
            _methodCASequential.UseMuxChannel = new BitArray(measureOnChannels1); //Set to measure sequentially on channel 1, 3 and 5

            //Alternating multiplexer measurement settings
            //Only Chronoamperometry (Amperometric Detection) and (Open Circuit) Potentiometry support alternating measurements. 
            _methodCAAlternating = new AmperometricDetection(); //Create a new chronoamperometry method with the default settings
            _methodCAAlternating.Potential = 1f; //Sets the potential at 1V
            _methodCAAlternating.IntervalTime = 0.25f; //Sets the sampling interval time at 0.25s (0.25s is the minimum sample time for the multiplexer in alternating mode)
            _methodCAAlternating.RunTime = 10f; //Sets the measurement runtime at 2.5s
            _methodCAAlternating.MuxMethod = MuxMethod.Alternatingly; //Set the type of multiplexer measurement
            //Create an array of booleans indicating on which channels you would like to measure
            //For a 16 channel mux you can create a bool array for 16 channels
            bool[] measureOnChannels2 = new bool[8];  //Define bool array for 8 channel multiplexer
            //Specify on which channels to measure
            //For alternating mux measurements only successive channels can be selected starting at the first channel (i.e. 1, 2 and 3 not for example 2, 4 and 7)                  
            measureOnChannels2[0] = true;
            measureOnChannels2[1] = true;
            measureOnChannels2[2] = true;
            measureOnChannels2[3] = true;
            measureOnChannels2[4] = true;
            measureOnChannels2[5] = true;
            measureOnChannels2[6] = false;
            measureOnChannels2[7] = false;
            _methodCAAlternating.UseMuxChannel = new BitArray(measureOnChannels2); //Set to measure alternatingly on channel 1, 2 and 3
        }

        /// <summary>
        /// Discovers the connected PalmSens & EmStat devices and adds them to the combobox control.
        /// </summary>
        private void DiscoverConnectedDevices()
        {
            cmbDevices.Items.Clear();
            _connectedDevices = psCommSimpleWPF.ConnectedDevices; //Discover connected devices

            foreach (Device d in _connectedDevices)
                cmbDevices.Items.Add(d.ToString()); //Add connected devices to control

            int nDevices = cmbDevices.Items.Count;
            cmbDevices.SelectedIndex = nDevices > 0 ? 0 : -1;
            lbConsole.Items.Add($"Found {nDevices} device(s).");

            btnConnect.IsEnabled = nDevices > 0;
        }

        /// <summary>
        /// Handles the Click event of the btnRefresh control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private void btnRefresh_Click(object sender, RoutedEventArgs e)
        {
            DiscoverConnectedDevices(); //Add connected devices to the devices combobox control
        }

        /// <summary>
        /// Handles the Click event of the btnConnect control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private void btnConnect_Click(object sender, RoutedEventArgs e)
        {
            if (!psCommSimpleWPF.Connected) //Determine whether a device is currently connected
            {
                if (cmbDevices.SelectedIndex == -1)
                    return;

                try
                {
                    //Connect to the device selected in the devices combobox control
                    psCommSimpleWPF.Connect(_connectedDevices[cmbDevices.SelectedIndex]);
                    lbConsole.Items.Add($"Connected to {psCommSimpleWPF.ConnectedDevice.ToString()}");
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            else
            {
                psCommSimpleWPF.Disconnect(); //Disconnect from the connected device
            }

            //Update UI based on connection status
            cmbDevices.IsEnabled = !psCommSimpleWPF.Connected;
            btnRefresh.IsEnabled = !psCommSimpleWPF.Connected;
            btnConnect.Content = psCommSimpleWPF.Connected ? "Disconnect" : "Connect";
            btnMeasureSingleChannel.IsEnabled = psCommSimpleWPF.Connected;
            btnMeasureSequentially.IsEnabled = psCommSimpleWPF.Connected;
            btnMeasureAlternatingly.IsEnabled = psCommSimpleWPF.Connected;
        }

        /// <summary>
        /// Handles the Click event of the btnMeasureSingleChannel control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private void btnMeasureSingleChannel_Click(object sender, RoutedEventArgs e)
        {
            if (psCommSimpleWPF.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    _dataPoints.Clear(); //Reset the data grid view control
                    //Start measurement defined in the method on the multiplexer channel specified in the channel combobox control
                    _activeMeasurement = psCommSimpleWPF.Measure(_methodCA, cmbChannels.SelectedIndex);
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            else
            {
                try
                {
                    psCommSimpleWPF.AbortMeasurement(); //Abort the active measurement
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
        }

        /// <summary>
        /// Handles the Click event of the btnMeasureSequentially control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private void btnMeasureSequentially_Click(object sender, RoutedEventArgs e)
        {
            if (psCommSimpleWPF.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    _dataPoints.Clear(); //Reset the data grid view control
                    //Start the sequential multiplexer measurement on the channels defined in the method
                    _activeMeasurement = psCommSimpleWPF.Measure(_methodCASequential);
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
        }

        /// <summary>
        /// Handles the Click event of the btnMeasureAlternatingly control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private void btnMeasureAlternatingly_Click(object sender, RoutedEventArgs e)
        {
            if (psCommSimpleWPF.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    _dataPoints.Clear(); //Reset the data grid view control
                    //Start the alternating multiplexer measurement on the channels defined in the method
                    _activeMeasurement = psCommSimpleWPF.Measure(_methodCAAlternating);
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
        }

        /// <summary>
        /// Raised when device status package is received (the device does not send status packages while measuring)
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="PalmSens.Comm.StatusEventArgs"/> instance containing the event data.</param>
        private void psCommSimpleWPF_ReceiveStatus(object sender, PalmSens.Comm.StatusEventArgs e)
        {
            Status status = e.GetStatus(); //Get the PalmSens.Comm.Status instance from the event data
            double potential = status.PotentialReading.Value; //Get the potential
            double currentInRange = status.CurrentReading.ValueInRange; //Get the current expressed inthe active current range
            PalmSens.Comm.ReadingStatus currentStatus = status.CurrentReading.ReadingStatus; //Get the status of the current reading
            CurrentRange cr = status.CurrentReading.CurrentRange; //Get the active current range

            tbPotential.Text = potential.ToString("F3");
            tbCurrent.Text = currentInRange.ToString("F3");
            lblCurrentRange.Content = $"* {cr.ToString()}";
        }

        /// <summary>
        /// Raised when the connected device's status changes
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="CurrentState">State of the current.</param>
        private void psCommSimpleWPF_StateChanged(object sender, PalmSens.Comm.CommManager.DeviceState CurrentState)
        {
            tbStatus.Text = CurrentState.ToString(); //Updates the device state indicator textbox
            btnConnect.IsEnabled = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle;
            btnMeasureSingleChannel.Content = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle ? "Measure" : "Abort";
            btnMeasureSequentially.IsEnabled = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle;
            btnMeasureAlternatingly.IsEnabled = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle;
        }

        /// <summary>
        /// Raised when the measurement is ended
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void psCommSimpleWPF_MeasurementEnded(object sender, Exception e)
        {
            lbConsole.Items.Add("Measurement ended.");
        }

        /// <summary>
        /// Raised when the measurement is started
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void psCommSimpleWPF_MeasurementStarted(object sender, EventArgs e)
        {
            lbConsole.Items.Add("Measurement started.");
        }

        /// <summary>
        /// Raised when a Simple Curve in the active SimpleMeasurement starts receiving data
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="activeSimpleCurve">The active simple curve.</param>
        private void psCommSimpleWPF_SimpleCurveStartReceivingData(object sender, SimpleCurve activeSimpleCurve)
        {
            //Subscribe to the curve's events to receive updates when new data is available and when it iss finished receiving data
            activeSimpleCurve.NewDataAdded += activeSimpleCurve_NewDataAdded;
            activeSimpleCurve.CurveFinished += activeSimpleCurve_CurveFinished;

            lbConsole.Items.Add($"Channel {activeSimpleCurve.MuxChannel + 1}: Curve is receiving new data...");
        }

        /// <summary>
        /// Raised when new data points are added to the active SimpleCurve
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="PalmSens.Data.ArrayDataAddedEventArgs"/> instance containing the event data.</param>
        private void activeSimpleCurve_NewDataAdded(object sender, PalmSens.Data.ArrayDataAddedEventArgs e)
        {
            SimpleCurve activeSimpleCurve = sender as SimpleCurve;
            int startIndex = e.StartIndex; //The index of the first new data point added to the curve
            int count = e.Count; //The number of new data points added to the curve

            for (int i = startIndex; i < startIndex + count; i++)
            {
                double xValue = activeSimpleCurve.XAxisValue(i); //Get the value on Curve's X-Axis (potential) at the specified index
                double yValue = activeSimpleCurve.YAxisValue(i); //Get the value on Curve's Y-Axis (current) at the specified index
                _dataPoints.Add(new DataPoint(i + 1, xValue, yValue, activeSimpleCurve.MuxChannel + 1));
            }
        }

        /// <summary>
        /// Raised when a SimpleCurve stops receiving new data points
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void activeSimpleCurve_CurveFinished(object sender, EventArgs e)
        {
            SimpleCurve activeSimpleCurve = sender as SimpleCurve;

            //Unsubscribe from the curves events to avoid memory leaks
            activeSimpleCurve.NewDataAdded -= activeSimpleCurve_NewDataAdded;
            activeSimpleCurve.CurveFinished -= activeSimpleCurve_CurveFinished;

            int nDataPointsReceived = activeSimpleCurve.NDataPoints;
            lbConsole.Items.Add($"Channel {activeSimpleCurve.MuxChannel + 1}: {nDataPointsReceived} data point(s) received.");
            lbConsole.Items.Add($"Channel {activeSimpleCurve.MuxChannel + 1}: Finished");
        }

        /// <summary>
        /// Raised when the instrument has been disconnected.
        /// If the instrument was disconnected due to a communication the exception is provided.
        /// In the case of a regular disconnect the exception will be set to null.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="exception">The exception.</param>
        private void psCommSimpleWPF_Disconnected(object sender, Exception exception)
        {
            if (exception != null)
            {
                lbConsole.Items.Add(exception.Message);
            }

            lbConsole.Items.Add("Disconnected.");
            btnConnect.Content = "Connect";
            btnConnect.IsEnabled = true;
            btnMeasureSingleChannel.Content = "Measure";
        }
    }

    /// <summary>
    /// Defines the row structure for the DataGrid items
    /// </summary>
    public class DataPoint
    {
        public int ID { get; }
        public double Potential { get; }
        public double Current { get; }
        public int MuxChannel { get; }

        public DataPoint(int id, double potential, double current, int channel)
        {
            ID = id;
            Potential = potential;
            Current = current;
            MuxChannel = channel;
        }
    }
}
