﻿using GalaSoft.MvvmLight.CommandWpf;
using Microsoft.Win32;
using PalmSens.Analysis;
using PalmSens.Core.Simplified.Data;
using PalmSens.Core.Simplified.WPF;
using SDKPlot.WPF;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Data;
using System.Windows.Input;

namespace PSSDKPeakBrowserExample
{
    public class ViewModel : INotifyPropertyChanged
    {
        public ViewModel()
        {
            PeakSettings = new PeakSettings
            {
                WindowLeft = 0f,
                WindowRight = 0f,
                MinHeight = 0f,
                MinWidth = 0f,
                Shoulder = false,
                Merge = false,
                Smooth = 0
            };
#if DEBUG
            LoadSession(Environment.CurrentDirectory + "\\DPV example.pssession");
#endif
        }

        public ViewModel(Plot plot) : this()
        {
            _plot = plot;
            UpdatePlot();
        }

        #region Properties
        private readonly Plot _plot;

        public readonly PeakSettings PeakSettings;

        private Dictionary<string, List<PeakData>> _loadedData = new Dictionary<string, List<PeakData>>();

        private List<MeasurementTypes> _compatibleMeasurementTypes = new List<MeasurementTypes>()
        {
            MeasurementTypes.CyclicVoltammetry,
            MeasurementTypes.LinearSweepVoltammetry,
            MeasurementTypes.ACVoltammetry,
            MeasurementTypes.DifferentialPulseVoltammetry,
            MeasurementTypes.NormalPulseVoltammetry,
            MeasurementTypes.SquareWaveVoltammetry
        };
        #endregion

        #region Binding
        public float WindowLeft
        {
            get { return PeakSettings.WindowLeft; }
            set
            {
                if (PeakSettings.WindowLeft == value) return;
                PeakSettings.WindowLeft = value;
                Update();
                OnPropertyChanged(nameof(WindowLeft));
            }
        }

        public float WindowRight
        {
            get { return PeakSettings.WindowRight; }
            set
            {
                if (PeakSettings.WindowRight == value) return;
                PeakSettings.WindowRight = value;
                Update();
                OnPropertyChanged(nameof(WindowRight));
            }
        }

        public float MinHeight
        {
            get { return PeakSettings.MinHeight; }
            set
            {
                if (PeakSettings.MinHeight == value) return;
                PeakSettings.MinHeight = value;
                Update();
                OnPropertyChanged(nameof(MinHeight));
            }
        }

        public float MinWidth
        {
            get { return PeakSettings.MinWidth; }
            set
            {
                if (PeakSettings.MinWidth == value) return;
                PeakSettings.MinWidth = value;
                Update();
                OnPropertyChanged(nameof(MinWidth));
            }
        }

        public int Smooth
        {
            get { return PeakSettings.Smooth; }
            set
            {
                if (PeakSettings.Smooth == value) return;
                PeakSettings.Smooth = value;
                Update();
                OnPropertyChanged(nameof(Smooth));
            }
        }

        public bool Shoulder
        {
            get { return PeakSettings.Shoulder; }
            set
            {
                if (PeakSettings.Shoulder == value) return;
                PeakSettings.Shoulder = value;
                Update();
                OnPropertyChanged(nameof(Shoulder));
            }
        }

        public bool Merge
        {
            get { return PeakSettings.Merge; }
            set
            {
                if (PeakSettings.Merge == value) return;
                PeakSettings.Merge = value;
                Update();
                OnPropertyChanged(nameof(Merge));
            }
        }

        private int _selectedCurveIndex;
        public int SelectedCurveIndex
        {
            get { return _selectedCurveIndex; }
            set
            {
                _selectedCurveIndex = value;
                UpdatePlot();
                OnPropertyChanged(nameof(SelectedCurveIndex));
            }
        }

        public ObservableCollection<string> SessionFiles { get; private set; } = new ObservableCollection<string>();

        public ObservableCollection<PeakData> PeakDatas { get; private set; } = new ObservableCollection<PeakData>();

        private RelayCommand _loadDataCommand;
        public ICommand LoadDataCommand { get { return _loadDataCommand ?? (_loadDataCommand = new RelayCommand(LoadData)); } }

        private RelayCommand<IList<object>> _deleteDataCommand;
        public ICommand DeleteDataCommand { get { return _deleteDataCommand ?? (_deleteDataCommand = new RelayCommand<IList<object>>((args) => DeleteData(args))); } }
        #endregion

        #region Functions
        public void LoadData()
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.Multiselect = true;
            openFileDialog.Filter = "PSSession files (*.pssession)|*.pssession";
            openFileDialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            if (openFileDialog.ShowDialog() == true)
                foreach (string filePath in openFileDialog.FileNames)
                    LoadSession(filePath);
            SelectedCurveIndex = PeakDatas.Count - 1;
            Update();
        }

        private void LoadSession(string filePath)
        {
            if (SessionFiles.Contains(filePath))
                return;

            List<SimpleMeasurement> measurements = null;

            try
            {
                measurements = SimpleLoadSaveFunctions.LoadMeasurements(filePath);
            }
            catch (Exception ex)
            {
                return;
            }

            List<PeakData> peakDatas = new List<PeakData>();

            foreach (SimpleMeasurement measurement in measurements)
                if (_compatibleMeasurementTypes.Contains(measurement.MeasurementType))
                    foreach (SimpleCurve curve in measurement.SimpleCurveCollection)
                        if ((curve.XAxisDataType == PalmSens.Data.DataArrayType.Potential || curve.Curve.XUnit is PalmSens.Units.Volt) &&
                            (curve.YAxisDataType == PalmSens.Data.DataArrayType.Current || curve.Curve.YUnit is PalmSens.Units.MicroAmpere))
                            peakDatas.Add(new PeakData(curve, measurement, filePath, PeakSettings));

            if (peakDatas.Count == 0)
                return;

            SessionFiles.Add(filePath);
            _loadedData.Add(filePath, peakDatas);
            foreach (PeakData peakData in peakDatas)
                PeakDatas.Add(peakData);
        }

        public void DeleteData(IList<object> selectedItems)
        {
            List<string> filesNames = selectedItems.Select(i => (string)i).ToList();
            foreach (string file in filesNames)
            {
                List<PeakData> curves = _loadedData[file];
                SessionFiles.Remove(file);
                _loadedData.Remove(file);
                foreach (PeakData data in curves)
                    PeakDatas.Remove(data);
            }
            Update();
        }

        public void Update()
        {
            foreach (PeakData peakData in PeakDatas)
                peakData.DetectPeaks();
            UpdatePlot();
        }

        private void UpdatePlot()
        {
            if (_plot == null) return;
            _plot.ClearAll();
            if (_selectedCurveIndex < PeakDatas.Count && _selectedCurveIndex > -1)
                _plot.AddSimpleCurve(PeakDatas[_selectedCurveIndex].ProcessedCurve);
        }
        #endregion

        #region Interface
        public event PropertyChangedEventHandler PropertyChanged;
        private void OnPropertyChanged(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }
        #endregion
    }

    public class PeakSettings
    {
        public float WindowLeft { get; set; }
        public float WindowRight { get; set; }
        public float MinHeight { get; set; }
        public float MinWidth { get; set; }
        public bool Shoulder { get; set; }
        public bool Merge { get; set; }
        public int Smooth { get; set; }
    }
}
